/*
 * Copyright (c) 2008-2020 Allwinner Technology Co. Ltd.
 * All rights reserved.
 *
 * File : ion_mem.c
 * Description :
 * History :
 *	Author	: jilinglin <jilinglin@allwinnertech.com>
 *	Date	: 2020/04/14
 *	Comment :
 */
#include "ion_mem.h"

#include <asm/cacheflush.h>
#include <linux/dma-buf.h>
#include <linux/dma-mapping.h>
#include <linux/ion.h>
#include <linux/slab.h>

#include "log.h"
#include "mem_interface.h"
#include "ve_mem_list.h"
// #include "cedar_ve_out.h"
extern struct device *cedar_get_device(void);
extern void cedar_dma_flush_range(const void *, size_t);
#define ION_SYSTEM_HEAP_MASK (1 << ION_HEAP_TYPE_SYSTEM)
#define DEBUG_SHOW_TOTAL_PALLOC_SIZE (0)
// #define CONFIG_SUNXI_IOMMU
struct buf_node {
	struct aw_mem_list_head entry;
	unsigned long phy_addr;	 // phisical address
	void *vir_addr;			 // virtual address
	unsigned int size;		 // buffer size
	int dma_buf_fd;
	struct dma_buf *buf;
	struct dma_buf_attachment *attach;
	struct sg_table *sgt;
};

struct cedar_cache_range {
	unsigned long start;
	unsigned long end;
};

struct mem_ion_ctx {
	struct mem_interface interface;
	int fd;						   // driver handle
	struct aw_mem_list_head list;  // buffer list
	int ref_cnt;				   // reference count
	unsigned int phyOffset;
	struct device *dev;
};

static int get_iommu_phy_addr(struct mem_ion_ctx *p, struct buf_node *node)
{
	struct sg_table *sgt;

	node->attach = dma_buf_attach(node->buf, p->dev);
	if (IS_ERR(node->attach)) {
		loge("dma_buf_attach failed\n");
		return -1;
	}
	sgt = dma_buf_map_attachment(node->attach, DMA_BIDIRECTIONAL);
	if (IS_ERR_OR_NULL(sgt)) {
		loge("dma_buf_map_attachment failed\n");
		goto err_buf_detach;
	}

	node->sgt = sgt;
	node->phy_addr = sg_dma_address(sgt->sgl);
	return 0;

err_buf_detach:
	dma_buf_detach(node->buf, node->attach);
	return -1;
}
#if DEBUG_SHOW_TOTAL_PALLOC_SIZE
static int total_palloc_size;
#endif

void *mem_ion_palloc_base(struct mem_interface *interface, int size,
						  int bIsCacheFlag)
{
	struct buf_node *node;
	unsigned int flags = 0;
	unsigned int heap_mask;
	unsigned int align;
	unsigned int buf_fd;

	struct dma_buf *dmabuf = NULL;
	struct mem_ion_ctx *p = (struct mem_ion_ctx *)interface;

	if (!p)
		return NULL;

	node = kmalloc(sizeof(struct buf_node), GFP_KERNEL);

	if (node == NULL) {
		logd("malloc buffer node fail");
		return NULL;
	}

	// flags = ION_FLAG_CACHED;
	// heap_mask = ION_HEAP_TYPE_DMA | ION_HEAP_TYPE_CARVEOUT;
#if IS_ENABLED(CONFIG_SUNXI_IOMMU)
	heap_mask = 1 << ION_HEAP_TYPE_SYSTEM;
#else
	heap_mask = 2; /*1 << ION_HEAP_TYPE_DMA;*/
#endif
	dmabuf = ion_alloc(size, heap_mask, flags);
	if (IS_ERR_OR_NULL(dmabuf)) {
		loge("%s: ion_alloc failed, size=%u dmabuf=0x%p\n", __func__,
			 (unsigned int)size, dmabuf);
		return -2;
	}

	node->vir_addr = ion_heap_map_kernel(NULL, dmabuf->priv);

	if (IS_ERR_OR_NULL(node->vir_addr)) {
		loge("ion_map_kernel failed");
		goto err_map;
	}

	buf_fd = dma_buf_fd(dmabuf, O_CLOEXEC);
	dmabuf = dma_buf_get(buf_fd);
	if (IS_ERR(dmabuf)) {
		loge("dma_buf_get failed, fd=%d\n", buf_fd);
		goto err_phyaddr;
	}
	node->size = size;
	node->buf = dmabuf;
	node->dma_buf_fd = buf_fd;

	if (get_iommu_phy_addr(p, node) != 0) {
		logd("iommu phys failed");
		goto err_phyaddr;
	}

	aw_mem_list_add_tail(&node->entry, &p->list);

#if DEBUG_SHOW_TOTAL_PALLOC_SIZE
	total_palloc_size += node->size;
	logd("palloc phy:%lx, vir:%08x, flag: %d, size:%d--%d kb, totalS:%d--%d kb",
		 node->phy_addr, (char *)node->vir_addr, flags, node->size,
		 node->size / 1024, total_palloc_size, total_palloc_size / 1024);
#endif
	return node->vir_addr;

err_phyaddr:
	ion_heap_unmap_kernel(NULL, dmabuf->priv);
err_map:
	dma_buf_put(dmabuf);
	put_unused_fd(buf_fd);
err_dmabuf:
	kfree(node);
	return NULL;
}

void *mem_ion_palloc(struct mem_interface *interface, int size)
{
	int bIsCacheFlag = 1;

	return mem_ion_palloc_base(interface, size, bIsCacheFlag);
}

void *mem_ion_palloc_no_cache(struct mem_interface *interface, int size)
{
	int bIsCacheFlag = 0;

	return mem_ion_palloc_base(interface, size, bIsCacheFlag);
}

void mem_ion_pfree(struct mem_interface *interface, void *pbuf)
{
	struct buf_node *tmp;

	struct mem_ion_ctx *p = (struct mem_ion_ctx *)interface;

	if (p == NULL) {
		loge(
			"ion_alloc do not opened, should call ion_alloc_open() \
			before ion_alloc_alloc(size) \n");
		return;
	}

	aw_mem_list_for_each_entry(tmp, &p->list, entry)
	{
		if (tmp->vir_addr == pbuf) {
#if DEBUG_SHOW_TOTAL_PALLOC_SIZE
			total_palloc_size -= tmp->size;
			logd("pfree phy1:%lx, vir:%p, size:%d--%d kb, totalS:%d--%d kb",
				 tmp->phy_addr, tmp->vir_addr, tmp->size, tmp->size / 1024,
				 total_palloc_size, total_palloc_size / 1024);
#endif
			dma_buf_unmap_attachment(tmp->attach, tmp->sgt, DMA_BIDIRECTIONAL);
			dma_buf_detach(tmp->buf, tmp->attach);
			dma_buf_put(tmp->buf);
			put_unused_fd(tmp->dma_buf_fd);

			ion_heap_unmap_kernel(NULL, tmp->buf->priv);
			ion_free(tmp->buf->priv);
			// put_unused_fd(tmp->dma_buf_fd);
			aw_mem_list_del(&tmp->entry);
			kfree(tmp);
			break;
		}
	}
}

void *mem_ion_phy2vir(struct mem_interface *interface, unsigned long addr_phy)
{
	int flag = 0;
	void *addr_vir = 0;
	struct buf_node *tmp;

	struct mem_ion_ctx *p = (struct mem_ion_ctx *)interface;

	aw_mem_list_for_each_entry(tmp, &p->list, entry)
	{
		if (addr_phy >= tmp->phy_addr && addr_phy < tmp->phy_addr + tmp->size) {
			addr_vir = tmp->vir_addr + addr_phy - tmp->phy_addr;
			flag = 1;
			break;
		}
	}

	if (0 == flag) {
		loge("ion_alloc_phy2vir failed, do not find physical address: 0x%lx \n",
			 addr_phy);
	}

	return addr_vir;
}

unsigned long mem_ion_vir2phy(struct mem_interface *interface, void *addr_vir)
{
	int flag = 0;
	unsigned long addr_phy = 0;
	struct buf_node *tmp;

	struct mem_ion_ctx *p = (struct mem_ion_ctx *)interface;

	aw_mem_list_for_each_entry(tmp, &p->list, entry)
	{
		// logd("vir:%p node vir:%p size:%d", addr_vir, tmp->vir_addr,
		// tmp->size);
		if (addr_vir >= tmp->vir_addr && addr_vir < tmp->vir_addr + tmp->size) {
			addr_phy = tmp->phy_addr + addr_vir - tmp->vir_addr;
			flag = 1;
			break;
		}
	}

	if (flag == 0)
		loge("ion_alloc_vir2phy failed, do not find physical address: %p \n",
			 addr_vir);

	return addr_phy;
}

void mem_ion_flush_cache(struct mem_interface *p, void *startAddr, int size)
{
	struct cedar_cache_range range;

	range.start = (unsigned long)startAddr;
	range.end = (unsigned long)startAddr + size;

	// logd("flush cache start:%lu end:%lu size:%d",range.start,range.end,size);
#if IS_ENABLED(CONFIG_ARM64)
	cedar_dma_flush_range((void *)range.start, (void *)range.end);
	// loge("cedar_dma_flush_range");
#else
	dmac_flush_range((void *)range.start, (void *)range.end);
	// loge("dmac_flush_range");
#endif

	return;
}

static int mem_ion_get_flag(struct mem_interface *p)
{
#if defined(CONFIG_SUNXI_IOMMU)
	return MEM_IOMMU_FLAG;
#else
	return MEM_NORMAL_FLAG;
#endif
}

int mem_ion_share_fd(struct mem_interface *interface, void *addr_vir)
{
	int flag = 0;
	struct buf_node *tmp;
	int share_fd = -1;

	struct mem_ion_ctx *p = (struct mem_ion_ctx *)interface;

	loge("not support!!!!");
	return -1;
	aw_mem_list_for_each_entry(tmp, &p->list, entry)
	{
		// logd("vir:%p node vir:%p size:%d", addr_vir, tmp->vir_addr,
		// tmp->size);
		if (addr_vir >= tmp->vir_addr && addr_vir < tmp->vir_addr + tmp->size) {
			// share_fd = ion_share_dma_buf_fd2(p->client, tmp->handle);
			flag = 1;
			break;
		}
	}

	if (flag == 0)
		loge("mem_ion_share_fd failed, do not find physical address: %p \n",
			 addr_vir);

	logv("share_fd = %d", share_fd);

	return share_fd;
}

struct mem_interface *mem_ion_create(void)
{
	struct mem_ion_ctx *context;

	context = kmalloc(sizeof(struct mem_ion_ctx), GFP_KERNEL);
	if (context == NULL) {
		loge("create ion allocator failed, out of memory \n");
		return NULL;
	}

	memset(context, 0, sizeof(struct mem_ion_ctx));

	context->interface.palloc_no_cache = mem_ion_palloc_no_cache;
	context->interface.palloc = mem_ion_palloc;
	context->interface.pfree = mem_ion_pfree;
	context->interface.get_phyaddr = mem_ion_vir2phy;
	context->interface.get_viraddr = mem_ion_phy2vir;
	context->interface.flush_cache = mem_ion_flush_cache;
	context->interface.get_mem_flag = mem_ion_get_flag;
	context->interface.share_fd = mem_ion_share_fd;

	AW_MEM_INIT_LIST_HEAD(&context->list);

	context->dev = cedar_get_device();

	return &context->interface;
}

void mem_ion_destory(struct mem_interface *p)
{
	struct mem_ion_ctx *context = (struct mem_ion_ctx *)p;

	if (context == NULL) {
		logd("you destroy mem but p is null");
		return;
	}

	kfree(context);
}
