
/*
* Copyright (c) 2008-2016 Allwinner Technology Co. Ltd.
* All rights reserved.
*
* File : mjpeg.c
* Description :
* History :
*   Author  : xyliu <xyliu@allwinnertech.com>
*   Date    : 2016/04/13
*   Comment :
*
*
*/

#include <linux/slab.h>
#include "mjpeg_dec.h"
#include "mjpeg_context.h"
#include "log.h"

volatile u64 uMjpegRegisterBaseAddr = 0;

//************************************************************************************/
//      Name:        MjpegDecoderInit                                                          //
//      Prototype:    Handle ve_open (VConfig* config, VideoStreamInfo* stream_info); //
//      Function:    Start up the VE CSP.                                             //
//      Return:    A handle of the VE device.                                           //
//      Input:    VConfig* config, the configuration for the VE CSP.                 //
//************************************************************************************/

static int  MjpegDecoderInit(DecoderInterface* pSelf,
            VConfig* pConfig, VideoStreamInfo* pVideoInfo, VideoFbmInfo* pFbmInfo)
{
    //*just for remove warning of cppcheck

    MjpegDecodeContext* pMjpegContext   = NULL;
    JpegDec* pMjpegDec = NULL;

    pMjpegContext = (MjpegDecodeContext*)pSelf;

    pMjpegContext->pFbmInfo = pFbmInfo;
    memcpy(&pMjpegContext->vconfig, pConfig, sizeof(VConfig));

    if(pMjpegContext->vconfig.nAlignStride == 32)
    {
        pMjpegContext->vconfig.nAlignStride = 16;
       //forced mjpeg422. mjpeg444 to mjpeg420, the alignstride only support 16
    }

    pMjpegContext->videoStreamInfo.eCodecFormat            = pVideoInfo->eCodecFormat;
    pMjpegContext->videoStreamInfo.nWidth                  = pVideoInfo->nWidth;
    pMjpegContext->videoStreamInfo.nHeight                 = pVideoInfo->nHeight;
    pMjpegContext->videoStreamInfo.nFrameRate              = pVideoInfo->nFrameRate;
    pMjpegContext->videoStreamInfo.nFrameDuration          = pVideoInfo->nFrameDuration;
    pMjpegContext->videoStreamInfo.nAspectRatio            = pVideoInfo->nAspectRatio;
    pMjpegContext->videoStreamInfo.nCodecSpecificDataLen   = pVideoInfo->nCodecSpecificDataLen;
    pMjpegContext->videoStreamInfo.pCodecSpecificData      = pVideoInfo->pCodecSpecificData;
    pMjpegContext->videoStreamInfo.bIs3DStream             = pVideoInfo->bIs3DStream;

    if(pMjpegContext->vconfig.bDisable3D == 1)
    {
        pMjpegContext->videoStreamInfo.bIs3DStream = 0;
    }

    pMjpegDec = kmalloc(sizeof(JpegDec), GFP_KERNEL | __GFP_ZERO);
    if(pMjpegDec == NULL)
    {
        goto _mjpeg_decode_init_err;
    }
    memset(pMjpegDec, 0, sizeof(JpegDec));
    uMjpegRegisterBaseAddr = (u64)ve_get_group_reg_addr(pMjpegContext->vconfig.veOpsS,
                                                             REG_GROUP_MPEG_DECODER);
    u32 nHuffmanTableSramBufLen = 1024;
    u32 nHuffmanTableSramBufSize = nHuffmanTableSramBufLen*sizeof(u16);
    pMjpegDec->pHuffmanTableSramBuf = (u16*)kmalloc(nHuffmanTableSramBufSize, GFP_KERNEL | __GFP_ZERO);
    if(pMjpegDec->pHuffmanTableSramBuf == NULL)
    {
        loge("malloc pHuffmanTableSramBuf failed, size = %d",nHuffmanTableSramBufSize);
        goto _mjpeg_decode_init_err;
    }
    memset(pMjpegDec->pHuffmanTableSramBuf, 0, nHuffmanTableSramBufSize);
    pMjpegContext->pMjpegDec = (void*)pMjpegDec;
    return VDECODE_RESULT_OK;
_mjpeg_decode_init_err:
    if(pMjpegDec != NULL)
    {
        kfree(pMjpegDec);
        pMjpegDec = NULL;
    }
    return VDECODE_RESULT_UNSUPPORTED;
}

/**********************************************************************   *******************/
//          Name:    ve_reset                                 /
//          Prototype: vresult_e ve_reset (u8 flush_pictures) //
//          Function:  Reset the VE CSP.                      //
//          Return:    VDECODE_RESULT_OK: success.            //
//                       VDECODE_RESULT_UNSUPPORTED: the CSP is not opened yet. //
/********************************************************************************************/

void MjpegDecoderReset(DecoderInterface* pSelf)
{
    MjpegDecodeContext* pMjpegContext = NULL;
    pMjpegContext = (MjpegDecodeContext*)pSelf;

    if(pMjpegContext == NULL)
    {
        return;
    }
    else
    {
        JpegDec* pMjpegDec = (JpegDec*)pMjpegContext->pMjpegDec;
        ve_reset(pMjpegContext->vconfig.veOpsS);
        pMjpegDec->nDecStep = MJPEG_DEC_INIT;
        pMjpegDec->nDecStep = MJPEG_DEC_INIT;
        return;
    }
}

/***********************************************************************/
//    Name:    ve_set_vbv                                                              //
//    Prototype: vresult_e ve_set_vbv (u8* vbv_buf, u32 nStreamBufferSize, Handle ve); //
//    Function: Set pSbm's bitstream buffer base address and buffer size to the CSP..  //
//    Return:   VDECODE_RESULT_OK: success.                                            //
//               VDECODE_RESULT_UNSUPPORTED: the CSP is not opened yet.                //
/*******************************************************************/

static int  MjpegDecoderSetSbm(DecoderInterface* pSelf, SbmInterface* pSbm, int nIndex)
{
    MjpegDecodeContext* pMjpegContext = NULL;
    pMjpegContext = (MjpegDecodeContext*)pSelf;

    if(pMjpegContext == NULL)
    {
        return VDECODE_RESULT_UNSUPPORTED;
    }
    else
    {
        if(nIndex == 0)
        {
            pMjpegContext->pSbm = pSbm;
            pMjpegContext->pVbvBase = (u8*)SbmBufferAddress(pSbm);
            pMjpegContext->nVbvSize = (u32)SbmBufferSize(pSbm);
        }
        else if(nIndex == 1)
        {
             pMjpegContext->pMinorSbm = pSbm;
            pMjpegContext->pMinorVbvBase = (u8*)SbmBufferAddress(pSbm);
            pMjpegContext->nMinorVbvSize = (u32)SbmBufferSize(pSbm);
        }
        return VDECODE_RESULT_OK;
    }
}

static int  MjpegDecoderGetFbmNum(DecoderInterface* pSelf)
{
    MjpegDecodeContext* pMjpegContext;

    pMjpegContext = (MjpegDecodeContext*)pSelf;

    if(pMjpegContext == NULL)
    {
        return 0;
    }
    if(pMjpegContext->videoStreamInfo.bIs3DStream==1)
    {
        return 2;
    }
    return 1;
}

/**********************************************************************/
//   Name:    Mpeg2DecoderGetFbm                                      //
//   Prototype: Handle get_fbm (void);                                //
//   Function: Get a handle of the pFbm instance,                     //
//             in which pictures for display are stored.              //
//   Return:   Not NULL Handle: handle of the pFbm instance.          //
//             NULL Handle: pFbm module is not initialized yet.       //
/***********************************************************************/

static Fbm* MjpegDecoderGetFbm(DecoderInterface* pSelf, int nIndex)
{
    MjpegDecodeContext* pMjpegContext = NULL;
    pMjpegContext = (MjpegDecodeContext*)pSelf;

    if(pMjpegContext == NULL)
    {
        return NULL;
    }
    else
    {
        if(nIndex == 0)
        {
            if(pMjpegContext->pFbmScaleDown != NULL)
            {
                return pMjpegContext->pFbmScaleDown;
            }
            else if(pMjpegContext->pFbm != NULL)
            {
                return pMjpegContext->pFbm;
            }
        }
        else if(nIndex == 1)
        {
            if(pMjpegContext->pMinorFbmScaleDown != NULL)
            {
                return pMjpegContext->pMinorFbmScaleDown;
            }
            else if(pMjpegContext->pMinorFbm != NULL)
            {
                return pMjpegContext->pMinorFbm;
            }
        }
    }
    return NULL;
}

/******************************************************************************************/
//       Name:    ve_close                                                                  //
//       Prototype: vresult_e ve_close (u8 flush_pictures);                               //
//       Function:    Close the VE CSP.                                                     //
//       Return:    VDECODE_RESULT_OK: success.                                                  //
//         VDECODE_RESULT_UNSUPPORTED: the CSP is not opened yet.                         //
/******************************************************************************************/

void DestroyMjpegDecoder(DecoderInterface* pSelf)
{
    MjpegDecodeContext* pMjpegContext = NULL;

    pMjpegContext = (MjpegDecodeContext*)pSelf;

    if(pMjpegContext == NULL)
    {
        return;
    }
    else
    {
        if(pMjpegContext->pFbm)
        {
            FbmDestroy(pMjpegContext->pFbm);
            pMjpegContext->pFbm = NULL;
        }
        if(pMjpegContext->pFbmScaleDown)
        {
            FbmDestroy(pMjpegContext->pFbmScaleDown);
            pMjpegContext->pFbmScaleDown = NULL;
        }
        if(pMjpegContext->videoStreamInfo.bIs3DStream == 1)
        {
            if(pMjpegContext->pMinorFbm)
                {
                FbmDestroy(pMjpegContext->pMinorFbm);
                pMjpegContext->pMinorFbm = NULL;
            }
            if(pMjpegContext->pMinorFbmScaleDown)
            {
                    FbmDestroy(pMjpegContext->pMinorFbmScaleDown);
                    pMjpegContext->pMinorFbmScaleDown = NULL;
            }
        }
        if(pMjpegContext->pMjpegDec != NULL)
        {
            JpegDec*pMjpegDec = (JpegDec*)pMjpegContext->pMjpegDec;
            if(pMjpegDec->pHuffmanTableSramBuf != NULL)
            {
                kfree(pMjpegDec->pHuffmanTableSramBuf);
                pMjpegDec->pHuffmanTableSramBuf = NULL;
            }

            kfree(pMjpegContext->pMjpegDec);
            pMjpegContext->pMjpegDec = NULL;
            pMjpegDec =NULL;
        }

        kfree(pMjpegContext);
        pMjpegContext = NULL;
        return;
    }
}


/*******************************************************************************/
//   Name:    ve_decode
//   Prototype: vresult_e decode (vstream_data_t* stream, u8 keyframe_only,
//              u8 bSkipBFrameIfDelay, u32 nCurrentTimeUs);
//   Function: Decode one bitstream frame.
//   INput:    vstream_data_t* stream: start address of the stream frame.
//             u8 keyframe_only:   tell the CSP to decode key frame only;
//             u8 bSkipBFrameIfDelay:  tell the CSP to skip B frame if it is overtime;
//             u32 nCurrentTimeUs:current time, used to compare with PTS when decoding B frame;
//   Return:   VDECODE_RESULT_OK:             decode stream success but no frame decoded;
//             VDECODE_RESULT_FRAME_DECODED:  one common frame decoded;
//             VRESULT_KEYFRAME_DECODED:    one key frame decoded;
//             VDECODE_RESULT_UNSUPPORTED:           decode stream fail;
//             VRESULT_ERR_INVALID_PARAM:  either stream or ve is NULL;
//             VRESULT_ERR_INVALID_STREAM: some error data in the stream, decode fail;
//             VRESULT_ERR_NO_MEMORY:      allocate memory fail in this method;
//             VRESULT_ERR_NO_FRAMEBUFFER: request empty frame buffer fail in this method;
//             VRESULT_ERR_UNSUPPORTED:    stream format is unsupported by this version of VE CSP;
//             VDECODE_RESULT_UNSUPPORTED:  'open' has not been successfully called yet.
/*****************************************************************************/

static int  MjpegDecoderDecode(DecoderInterface* pSelf,
                            int               bEndOfStream,
                            int               bDecodeKeyFrameOnly,
                            int               bSkipBFrameIfDelay,
                            int64_t           nCurrentTimeUs)
{
    s32 ret = 0;
    u8  buffer[4];
    u32 frameSize = 0;
    u32 remainData = 0;
    u32 nInLumaBufAddr = 0;
    u32 nInChromaBufAddr = 0;
    u32 nOutLumaBufAddr = 0;
    u32 nOutChromaBufAddr = 0;
    VideoStreamDataInfo* stream = NULL;
    JpegDec*        pMjpegDec = NULL;
    MjpegDecodeContext* pMjpegContext = NULL;
    u8  bDispFrmFlag = 0;

    pMjpegContext = (MjpegDecodeContext*)pSelf;
    pMjpegDec = (JpegDec*)pMjpegContext->pMjpegDec;

    // step 1:

    if(pMjpegDec->nDecStep == MJPEG_DEC_INIT)
    {
        if(pMjpegDec->nDecStramIndex == 0)
        {
            pMjpegDec->pSbm = pMjpegContext->pSbm;
            pMjpegDec->pFbm = pMjpegContext->pFbm;
            pMjpegDec->pFbmScaleDown = pMjpegContext->pFbmScaleDown;
            pMjpegDec->pVbvBase = pMjpegContext->pVbvBase;
            pMjpegDec->nVbvSize = pMjpegContext->nVbvSize;
        }
        else if(pMjpegDec->nDecStramIndex == 1)
        {
            pMjpegDec->pSbm = pMjpegContext->pMinorSbm;
            pMjpegDec->pVbvBase = pMjpegContext->pMinorVbvBase;
            pMjpegDec->nVbvSize = pMjpegContext->nMinorVbvSize;
            pMjpegDec->pFbm = pMjpegContext->pMinorFbm;
            pMjpegDec->pFbmScaleDown = pMjpegContext->pMinorFbmScaleDown;
        }
        pMjpegDec->nDecStep = MJPEG_DEC_GET_FBM_BUFFER;
    }

    //step2
    if(pMjpegDec->nDecStep == MJPEG_DEC_GET_FBM_BUFFER)
    {
         if(pMjpegDec->pFbm && FbmEmptyBufferNum(pMjpegDec->pFbm) == 0)
         {
             return VDECODE_RESULT_NO_FRAME_BUFFER;
         }

         if(pMjpegDec->pFbmScaleDown)
         {
             u32 num = FbmEmptyBufferNum(pMjpegDec->pFbmScaleDown);
              if(num < 2)
              {
                  return VDECODE_RESULT_NO_FRAME_BUFFER;
              }
         }
         pMjpegDec->nDecStep = MJPEG_DEC_GET_STREAM;
    }

    //step3
    if(pMjpegDec->nDecStep == MJPEG_DEC_GET_STREAM)
    {
        u8* ptr = NULL;
        stream = SbmRequestStream(pMjpegDec->pSbm);
        if(stream == NULL)
        {
            return VDECODE_RESULT_NO_BITSTREAM;
        }
        pMjpegDec->nPts = stream->nPts;
        if((u64)(stream->pData+4) <= (u64)(pMjpegDec->pVbvBase+pMjpegDec->nVbvSize))
        {
            ptr = (u8*)(stream->pData);
        }
        else
        {
            remainData = ((u64)pMjpegDec->pVbvBase+pMjpegDec->nVbvSize-
                (u64)stream->pData);
            memcpy(buffer, (u8*)stream->pData, remainData);
            memcpy(buffer+remainData,  pMjpegDec->pVbvBase, 4-remainData);
            ptr = buffer;
        }
        frameSize = (ptr[0]<<24)|(ptr[1]<<16)|(ptr[2]<<8)|(ptr[3]);
        if(frameSize+4 == (u32)stream->nLength)
        {
            pMjpegDec->nFrameSize = frameSize;
            pMjpegDec->data = (u8*)(stream->pData+4);
            if(pMjpegDec->data > pMjpegDec->pVbvBase+pMjpegDec->nVbvSize)
            {
                pMjpegDec->data -= pMjpegDec->nVbvSize;
            }
        }
        else
        {
            pMjpegDec->nFrameSize = stream->nLength;
            pMjpegDec->data = (u8*)stream->pData;
        }
        pMjpegDec->nDecStep = MJPEG_DEC_STREAM_DATA;
    }

    //step4:
    if(pMjpegDec->nDecStep == MJPEG_DEC_STREAM_DATA)
    {
        //ve_reset(pMjpegContext->vconfig.veOpsS);
        ret = JpegDecoderMain(pMjpegContext, pMjpegDec);
        if(ret == VDECODE_RESULT_NO_FRAME_BUFFER)
        {
            if(stream != NULL)
                SbmReturnStream(pMjpegDec->pSbm, stream);
            return VDECODE_RESULT_NO_FRAME_BUFFER;
        }
        else if(ret < 0)
        {
            if(stream != NULL)
                SbmFlushStream(pMjpegDec->pSbm, stream);
            pMjpegDec->nDecStep = MJPEG_DEC_GET_STREAM;
            return VDECODE_RESULT_OK;
        }
        else if(pMjpegDec->enRotationOut == 1)
        {
            bDispFrmFlag = 1;
            pMjpegDec->nDecStep = MJPEG_DEC_PROCESS_ROTATE;
        }
        else
        {
            bDispFrmFlag = 1;
            pMjpegDec->nDecStep = MJPEG_DEC_PROCESS_RESULT;
        }
    }

    if(pMjpegDec->nDecStep == MJPEG_DEC_PROCESS_ROTATE)
    {
        u8* pInLumaBuf = NULL;
        if(pMjpegDec->pReconPicture != NULL)
        {
            nInLumaBufAddr = pMjpegDec->nReconPicLumaAddr;
            nInChromaBufAddr = pMjpegDec->nReconPicChromaAddr;
            pInLumaBuf = (u8*)pMjpegDec->pReconPicture->pData0;
        }
        else
        {
            nInLumaBufAddr = pMjpegDec->nRefPicLumaAddr;
            nInChromaBufAddr = pMjpegDec->nRefPicChromaAddr;
            pInLumaBuf = (u8*)pMjpegDec->pRefPicture->pData0;
        }

        pMjpegDec->pDispPicture =  FbmRequestBuffer(pMjpegDec->pFbmScaleDown);
        if(pMjpegDec->pDispPicture == NULL)
        {
            if(stream != NULL)
            {
                SbmReturnStream(pMjpegDec->pSbm, stream);
                stream = NULL;
            }

            return VDECODE_RESULT_NO_FRAME_BUFFER;
        }
        nOutLumaBufAddr = pMjpegDec->pDispPicture->phyYBufAddr;
        nOutChromaBufAddr= pMjpegDec->pDispPicture->phyCBufAddr;

        if(pMjpegDec->bYv12OutFlag==1 ||
            pMjpegDec->bNv21OutFlag==1  ||
            pMjpegDec->bRefDispSameFormat==1)
        {
            if(pMjpegDec->jpegFormat==JPEG422 && pMjpegDec->bHwForce420Out==0)
            {
                MjpegTransformPlanner422To420(pMjpegContext,pMjpegDec,pInLumaBuf);
            }
            RotateMjpegBySoft(pInLumaBuf, (u8*)pMjpegDec->pDispPicture->pData0,
                    pMjpegDec->rotationDegree,
                    pMjpegDec->nWidth>>pMjpegDec->horizontalScaleRatio,
                    pMjpegDec->nHeight>>pMjpegDec->verticalScaleRatio, pMjpegContext);

            if(pMjpegDec->bRefDispSameFormat==1)
            {
               pMjpegDec->pDispPicture->ePixelFormat = pMjpegDec->eDispFormat;
            }
            else if(pMjpegDec->bYv12OutFlag==1)
            {
                pMjpegDec->pDispPicture->ePixelFormat = PIXEL_FORMAT_YV12;
            }
            else
            {
                pMjpegDec->pDispPicture->ePixelFormat = PIXEL_FORMAT_NV21;
            }

        }
        else  // MB mode
        {
            RotatMjpegByHardware(pMjpegContext, nInLumaBufAddr,
                nInChromaBufAddr,nOutLumaBufAddr, nOutChromaBufAddr);
            pMjpegDec->pDispPicture->ePixelFormat = PIXEL_FORMAT_YUV_MB32_420;
        }
        pMjpegDec->nDecStep = MJPEG_DEC_PROCESS_RESULT;
    }

    if(pMjpegDec->nDecStep == MJPEG_DEC_PROCESS_RESULT)
    {
        pMjpegDec->nDecFrameCount++;

        if(stream != NULL)
            SbmFlushStream(pMjpegDec->pSbm, stream);

        if(pMjpegDec->pDispPicture != NULL)
        {
            MjpegSetDispParams(pMjpegContext, pMjpegDec, pMjpegDec->pDispPicture);

            FbmReturnBuffer(pMjpegDec->pFbmScaleDown, pMjpegDec->pDispPicture, bDispFrmFlag);

            FbmReturnBuffer(pMjpegDec->pFbm, pMjpegDec->pRefPicture, 0);
            if(pMjpegDec->pReconPicture != NULL)
            {
                FbmReturnBuffer(pMjpegDec->pFbmScaleDown, pMjpegDec->pReconPicture, 0);
            }
        }
        else if(pMjpegDec->pReconPicture != NULL)
        {
            MjpegSetDispParams(pMjpegContext, pMjpegDec, pMjpegDec->pReconPicture);
            FbmReturnBuffer(pMjpegDec->pFbmScaleDown, pMjpegDec->pReconPicture, bDispFrmFlag);

            FbmReturnBuffer(pMjpegDec->pFbm, pMjpegDec->pRefPicture, 0);
        }
        else if(pMjpegDec->pRefPicture != NULL)
        {
            MjpegSetDispParams(pMjpegContext, pMjpegDec, pMjpegDec->pRefPicture);
            FbmReturnBuffer(pMjpegDec->pFbm, pMjpegDec->pRefPicture, bDispFrmFlag);
        }

        pMjpegDec->pDispPicture = NULL;
        pMjpegDec->pReconPicture = NULL;
        pMjpegDec->pRefPicture = NULL;
        pMjpegDec->nDecStep = MJPEG_DEC_INIT;

        if(pMjpegContext->videoStreamInfo.bIs3DStream==1)
        {
            pMjpegDec->nDecStramIndex = 1-pMjpegDec->nDecStramIndex;
        }
    }
    return (bDispFrmFlag==1)? VDECODE_RESULT_KEYFRAME_DECODED:VDECODE_RESULT_OK;
}

DecoderInterface* CreateMjpegDecoder(void)
{
    MjpegDecodeContext* pMjpegContext;

    pMjpegContext = (MjpegDecodeContext*) kmalloc(sizeof(MjpegDecodeContext), GFP_KERNEL | __GFP_ZERO);
    if(pMjpegContext == NULL)
        return NULL;

    memset(pMjpegContext, 0, sizeof(MjpegDecodeContext));

    pMjpegContext->interface.Init      = MjpegDecoderInit;
    pMjpegContext->interface.Reset     = MjpegDecoderReset;
    pMjpegContext->interface.SetSbm    = MjpegDecoderSetSbm;
    pMjpegContext->interface.GetFbmNum = MjpegDecoderGetFbmNum;
    pMjpegContext->interface.GetFbm    = MjpegDecoderGetFbm;
    pMjpegContext->interface.Decode    = MjpegDecoderDecode;

    return &pMjpegContext->interface;
}
