
/*
* Copyright (c) 2008-2016 Allwinner Technology Co. Ltd.
* All rights reserved.
*
* File : mjpeg_dec.c
* Description :
* History :
*   Author  : xyliu <xyliu@allwinnertech.com>
*   Date    : 2016/04/13
*   Comment :
*
*
*/
#include <linux/slab.h>
#include "mjpeg_dec.h"
#include "log.h"

extern volatile u64 uMjpegRegisterBaseAddr;
extern s32 InitJpegHw(JpegDec* jpgctx, MjpegDecodeContext* pMjpegContext);
extern s32 SetJpegFormat(JpegDec* jpgctx);
extern s32 JpegHwDec(JpegDec* jpgctx, MjpegDecodeContext* pMjpegContext);

static int GetBufferSize(int ePixelFormat, int nWidth, int nHeight,
        int*nYBufferSize, int *nCBufferSize, int* nYLineStride,
        int* nCLineStride, int nAlignValue)
{
    int   nHeight16Align;
    int   nHeight32Align;
    int   nHeight64Align;
    int   nLineStride;
    int   nMemSizeY;
    int   nMemSizeC;

    nHeight16Align = (nHeight+15) & ~15;
    nHeight32Align = (nHeight+31) & ~31;
    nHeight64Align = (nHeight+63) & ~63;

    switch(ePixelFormat)
    {
        case PIXEL_FORMAT_YUV_PLANER_420:
        case PIXEL_FORMAT_YUV_PLANER_422:
        case PIXEL_FORMAT_YUV_PLANER_444:
        case PIXEL_FORMAT_YV12:
        case PIXEL_FORMAT_NV21:

            //* for decoder,
            //* height of Y component is required to be 16 aligned,
            //* for example, 1080 becomes to 1088.
            //* width and height of U or V component are both required to be 8 aligned.
            //* nLineStride should be 16 aligned.
            if(nAlignValue == 0)
            {
                nLineStride = (nWidth+15) &~ 15;
                nMemSizeY = nLineStride*nHeight16Align;
            }
            else
            {
                nLineStride = (nWidth+nAlignValue-1) &~ (nAlignValue-1);
                nHeight = (nHeight+nAlignValue-1) &~ (nAlignValue-1);
                nMemSizeY = nLineStride*nHeight;
            }
            if(ePixelFormat == PIXEL_FORMAT_YUV_PLANER_420 ||
               ePixelFormat == PIXEL_FORMAT_YV12 ||
               ePixelFormat == PIXEL_FORMAT_NV21)
                nMemSizeC = nMemSizeY>>2;
            else if(ePixelFormat == PIXEL_FORMAT_YUV_PLANER_422)
                nMemSizeC = nMemSizeY>>1;
            else
                nMemSizeC = nMemSizeY;  //* PIXEL_FORMAT_YUV_PLANER_444
            break;

        case PIXEL_FORMAT_YUV_MB32_420:
        case PIXEL_FORMAT_YUV_MB32_422:
        case PIXEL_FORMAT_YUV_MB32_444:
            //* for decoder,
            //* height of Y component is required to be 32 aligned.
            //* height of UV component are both required to be 32 aligned.
            //* nLineStride should be 32 aligned.
            nLineStride = (nWidth+63) &~ 63;
            nMemSizeY = nLineStride*nHeight32Align;

            if(ePixelFormat == PIXEL_FORMAT_YUV_MB32_420)
                nMemSizeC = nLineStride*nHeight64Align/4;
            else if(ePixelFormat == PIXEL_FORMAT_YUV_MB32_422)
                nMemSizeC = nLineStride*nHeight64Align/2;
            else
                nMemSizeC = nLineStride*nHeight64Align;
            break;

        case PIXEL_FORMAT_RGBA:
        case PIXEL_FORMAT_ARGB:
        case PIXEL_FORMAT_ABGR:
        case PIXEL_FORMAT_BGRA:
//            nLineStride = (nWidth+3) &~ 3;
            nLineStride = nWidth *4;

            nMemSizeY = nLineStride*nHeight;
            nMemSizeC = 0;

            break;

        default:
            loge("pixel format incorrect, ePixelFormat=%d", ePixelFormat);
            return -1;
    }
    if(nYBufferSize != NULL)
    {
        *nYBufferSize = nMemSizeY;
    }
    if(nCBufferSize != NULL)
    {
        *nCBufferSize = nMemSizeC;
    }
    if(nYLineStride != NULL)
    {
        *nYLineStride = nLineStride;
    }
    if(nCLineStride != NULL)
    {
        *nCLineStride = nLineStride>>1;
    }
    return 0;
}
static u32 GetByte(JpegDec* p, u32 nBits)
{
    u32 i;
    s32 ret = 0;
    for(i=0; i<nBits; i+=8)
    {
        ret <<= 8;
        ret |= *p->data++;
        if(p->data >= (p->pVbvBase + p->nVbvSize))
            p->data = p->pVbvBase;
        p->nFrameSize--;
    }

    return ret;
}

#define INPUT_BYTE(x)     (x=GetByte(jpgctx,1*8))
#define INPUT_2BYTES(x)    (x=GetByte(jpgctx,2*8))

static void JpegSkipVariable (JpegDec* jpgctx)
{
    s32 length;

    INPUT_2BYTES(length);
    length -= 2;

    if (length > 0)
    {
        s32 i;

        for(i=0;i<(length>>2);i++)
        {
            GetByte(jpgctx, 4*8);
        }

        for(i=0;i<(length&3);i++)
        {
            GetByte(jpgctx, 1*8);
        }
    }
}

s32 JpegFirstMarker(JpegDec* jpgctx)
{
    u8 c, c2;

    INPUT_BYTE(c);
    INPUT_BYTE(c2);
    if (c != 0xFF || c2 != M_SOI)
        return 0;

    jpgctx->unreadMarker = c2;

    return 1;
}

s32 JpegNextMarker(JpegDec* jpgctx)
{
    u8 c;

    for (;;)
    {
        INPUT_BYTE(c);
        while (c != 0xFF)
        {
            INPUT_BYTE(c);
        }
        do
        {
            INPUT_BYTE(c);
        } while (c == 0xFF);
        if (c != 0)
            break;            /* found a valid marker, exit loop */
    }
    jpgctx->unreadMarker = c;

    return 1;
}

static s32 JpegGetSoi(JpegDec* jpgctx)
{
    if (jpgctx->sawSOI)
        return 0;

    jpgctx->sawSOI = 1;

    return 1;
}
void MjpegComputeOffset(MjpegDecodeContext* pMjpegContext, JpegDec* jpgctx, u32 width, u32 height)
{
    u32 nCropBottom = 0;
    u32 nCropLeft    = 0;
    u32 nCropTop    = 0;
    u32 nCropRight    = 0;
    if((pMjpegContext->vconfig.bRotationEn == 0)&&(pMjpegContext->vconfig.bScaleDownEn == 0))
    {
        nCropBottom = jpgctx->pFbm->pFrames->vpicture.nHeight-height;
        nCropLeft   = jpgctx->nLeftOffset;
        nCropTop    = jpgctx->nTopOffset;
        nCropRight  = jpgctx->pFbm->pFrames->vpicture.nWidth-width;

        jpgctx->nLeftOffset    = nCropLeft;
        jpgctx->nTopOffset     = nCropTop;
        jpgctx->nBottomOffset  = jpgctx->pFbm->pFrames->vpicture.nHeight-nCropBottom;
        jpgctx->nRightOffset   = jpgctx->pFbm->pFrames->vpicture.nWidth-nCropRight;
    }
    if((pMjpegContext->vconfig.nRotateDegree==2)
        ||(pMjpegContext->vconfig.nVerticalScaleDownRatio!=0)
        ||(pMjpegContext->vconfig.nHorizonScaleDownRatio!=0))
    {
        nCropBottom = jpgctx->pFbmScaleDown->pFrames->vpicture.nHeight-height;
        nCropLeft   = jpgctx->nLeftOffset;
        nCropTop    = jpgctx->nTopOffset;
        nCropRight  = jpgctx->pFbmScaleDown->pFrames->vpicture.nWidth-width;
    }
    if((pMjpegContext->vconfig.nRotateDegree==1)||(pMjpegContext->vconfig.nRotateDegree==3))
    {
        nCropLeft   = jpgctx->pFbmScaleDown->pFrames->vpicture.nHeight-height;
        nCropTop    = jpgctx->nLeftOffset;
        nCropRight  = jpgctx->nTopOffset;
        nCropBottom = jpgctx->pFbmScaleDown->pFrames->vpicture.nWidth-width;
    }

    if(pMjpegContext->vconfig.nRotateDegree==1)
    {
        jpgctx->nTopOffset     = nCropLeft;
        jpgctx->nLeftOffset    = nCropBottom;
        jpgctx->nRightOffset   = jpgctx->pFbmScaleDown->pFrames->vpicture.nWidth-nCropTop;
        jpgctx->nBottomOffset  = jpgctx->pFbmScaleDown->pFrames->vpicture.nHeight-nCropRight;
    }
    else if(pMjpegContext->vconfig.nRotateDegree == 2)
    {
        jpgctx->nTopOffset     = nCropBottom;
        jpgctx->nLeftOffset    = nCropRight;
        jpgctx->nRightOffset   = jpgctx->pFbmScaleDown->pFrames->vpicture.nWidth-nCropLeft;
        jpgctx->nBottomOffset  = jpgctx->pFbmScaleDown->pFrames->vpicture.nHeight-nCropTop;
    }
    else if(pMjpegContext->vconfig.nRotateDegree==3)
    {
        jpgctx->nTopOffset      = nCropRight;
        jpgctx->nLeftOffset     = nCropTop;
        jpgctx->nRightOffset    = jpgctx->pFbmScaleDown->pFrames->vpicture.nWidth-nCropBottom;
        jpgctx->nBottomOffset   = jpgctx->pFbmScaleDown->pFrames->vpicture.nHeight-nCropLeft;
    }
    if((pMjpegContext->vconfig.bRotationEn == 0)
        &&((pMjpegContext->vconfig.nVerticalScaleDownRatio!=0)
        ||(pMjpegContext->vconfig.nHorizonScaleDownRatio!=0)))
    {
        jpgctx->nLeftOffset    = nCropLeft;
        jpgctx->nTopOffset     = nCropTop;
        jpgctx->nBottomOffset  = jpgctx->pFbmScaleDown->pFrames->vpicture.nHeight-nCropBottom;
        jpgctx->nRightOffset   = jpgctx->pFbmScaleDown->pFrames->vpicture.nWidth-nCropRight;
    }
}
s32 JpegMallocFrmBuffer(MjpegDecodeContext* pMjpegContext, JpegDec* jpgctx)
{
    u32 width = 0;
    u32 height = 0;
    //u32 jpegFormat = 0;
    //u32 lineStride = 0;
    u32 nRefPixelFormat = 0;
    u32 nDispPixelFormat = 0;
    u8  JpegFormat[6] = {PIXEL_FORMAT_YUV_MB32_420, PIXEL_FORMAT_YUV_MB32_422,
                        PIXEL_FORMAT_YUV_MB32_422, PIXEL_FORMAT_YUV_MB32_422,
                         PIXEL_FORMAT_YUV_MB32_422, PIXEL_FORMAT_YUV_MB32_420};
    s32 nAlignValue = 0;
    s32 nProgressiveFlag = 1;
    FbmCreateInfo mFbmCreateInfo;
    VideoFbmInfo *pFbmInfo;

    #define MJPEG_FRM_BUF_NUM 8

    width = jpgctx->nWidth;
    height = jpgctx->nHeight;

    if(pMjpegContext->vconfig.nRotateDegree > 0)
    {
        pMjpegContext->vconfig.bRotationEn = 1;
    }

    jpgctx->nVeVersion =  ve_get_ic_version(pMjpegContext->vconfig.veOpsS);

    nRefPixelFormat = JpegFormat[jpgctx->jpegFormat];
    nDispPixelFormat = JpegFormat[jpgctx->jpegFormat];

    if(jpgctx->nVeVersion >=0x1639)
    {
        jpgctx->bHwForce420Out = 1;
        nDispPixelFormat = PIXEL_FORMAT_YUV_MB32_420;
    }
    jpgctx->bRefDispSameFormat = 0;

    if(jpgctx->nVeVersion >= 0x1667)
    {
        if(pMjpegContext->vconfig.eOutputPixelFormat == PIXEL_FORMAT_YV12)
        {
            nRefPixelFormat = PIXEL_FORMAT_YV12;
            nDispPixelFormat = PIXEL_FORMAT_YV12;
            jpgctx->eDispFormat = DISP_FORMAT_YV12;
            jpgctx->bRefDispSameFormat = 1;
        }
        else if(pMjpegContext->vconfig.eOutputPixelFormat ==
            PIXEL_FORMAT_YUV_PLANER_420)
        {
            nRefPixelFormat = PIXEL_FORMAT_YUV_PLANER_420;
            nDispPixelFormat = PIXEL_FORMAT_YUV_PLANER_420;
            jpgctx->eDispFormat = DISP_FORMAT_YU12;
            jpgctx->bRefDispSameFormat = 1;
        }
        else if(pMjpegContext->vconfig.eOutputPixelFormat == PIXEL_FORMAT_NV21)
        {
            nRefPixelFormat = PIXEL_FORMAT_NV21;
            nDispPixelFormat = PIXEL_FORMAT_NV21;
            jpgctx->eDispFormat = DISP_FORMAT_NV21;
            jpgctx->bRefDispSameFormat = 1;
        }
        else if(pMjpegContext->vconfig.eOutputPixelFormat == PIXEL_FORMAT_NV12)
        {
            nRefPixelFormat = PIXEL_FORMAT_NV12;
            nDispPixelFormat = PIXEL_FORMAT_NV12;
            jpgctx->eDispFormat = DISP_FORMAT_NV12;
            jpgctx->bRefDispSameFormat = 1;
        }
    }
    else
    {
        if(pMjpegContext->vconfig.eOutputPixelFormat == PIXEL_FORMAT_YV12)
        {
            jpgctx->bYv12OutFlag = 1;
            nDispPixelFormat = PIXEL_FORMAT_YV12;
            nRefPixelFormat = JpegFormat[jpgctx->jpegFormat];
            if(jpgctx->nVeVersion >=0x1639)
            {
                jpgctx->bHwForce420Out = 1;
                nRefPixelFormat = PIXEL_FORMAT_YUV_MB32_420;
            }
        }
        else if(pMjpegContext->vconfig.eOutputPixelFormat == PIXEL_FORMAT_NV21)
        {
            jpgctx->bNv21OutFlag = 1;
            nDispPixelFormat = PIXEL_FORMAT_NV21;
            nRefPixelFormat = JpegFormat[jpgctx->jpegFormat];
            if(jpgctx->nVeVersion >=0x1639)
            {
                jpgctx->bHwForce420Out = 1;
                   nRefPixelFormat = PIXEL_FORMAT_YUV_MB32_420;
            }
        }
    }

    jpgctx->enRotationOut         =  pMjpegContext->vconfig.bRotationEn;
    jpgctx->enScaleDownOut        =  pMjpegContext->vconfig.bScaleDownEn;
    jpgctx->verticalScaleRatio    = pMjpegContext->vconfig.nVerticalScaleDownRatio;
    jpgctx->horizontalScaleRatio  = pMjpegContext->vconfig.nHorizonScaleDownRatio;
    jpgctx->rotationDegree        = pMjpegContext->vconfig.nRotateDegree;

    if(jpgctx->rotationDegree==1 || jpgctx->rotationDegree==3)
    {
        jpgctx->verticalScaleRatio = pMjpegContext->vconfig.nHorizonScaleDownRatio;
        jpgctx->horizontalScaleRatio =  pMjpegContext->vconfig.nVerticalScaleDownRatio;
    }

    jpgctx->pFbm = NULL;
    jpgctx->pFbmScaleDown = NULL;
    pFbmInfo = (VideoFbmInfo *)pMjpegContext->pFbmInfo;

    if((pMjpegContext->vconfig.bScaleDownEn==0) &&
        (pMjpegContext->vconfig.bRotationEn==0) &&
        (jpgctx->bYv12OutFlag==0) &&
        (jpgctx->bNv21OutFlag==0))
    {
        memset(&mFbmCreateInfo, 0, sizeof(FbmCreateInfo));
        mFbmCreateInfo.nDecoderNeededMiniFrameNum = pFbmInfo->nDecoderNeededMiniFbmNum;
        mFbmCreateInfo.nFrameNum          =
            pFbmInfo->nDecoderNeededMiniFbmNum + pFbmInfo->nExtraFbmBufferNum;
        mFbmCreateInfo.nWidth             = width;
        mFbmCreateInfo.nHeight            = height;
        mFbmCreateInfo.ePixelFormat       = nRefPixelFormat;
        mFbmCreateInfo.bThumbnailMode     = pMjpegContext->vconfig.bThumbnailMode;
        //mFbmCreateInfo.callback           = pMjpegContext->vconfig.callback;
        //mFbmCreateInfo.pUserData          = pMjpegContext->vconfig.pUserData;
        mFbmCreateInfo.bGpuBufValid       = pMjpegContext->vconfig.bGpuBufValid;
        mFbmCreateInfo.nAlignStride       = pMjpegContext->vconfig.nAlignStride;
        mFbmCreateInfo.nBufferType        = BUF_TYPE_REFERENCE_DISP;
        mFbmCreateInfo.bProgressiveFlag   = nProgressiveFlag;
        mFbmCreateInfo.bIsSoftDecoderFlag = pMjpegContext->vconfig.bIsSoftDecoderFlag;
        mFbmCreateInfo.memops             = pMjpegContext->vconfig.memops;
        mFbmCreateInfo.veOpsS             = pMjpegContext->vconfig.veOpsS;

        jpgctx->pFbm = FbmCreate(&mFbmCreateInfo, pMjpegContext->pFbmInfo);

        if(jpgctx->pFbm == NULL)
        {
            return VDECODE_RESULT_UNSUPPORTED;
        }
        nAlignValue =  FbmGetAlignValue(jpgctx->pFbm);
        GetBufferSize((int)nRefPixelFormat,(int)width,(int)height,
                        (int*)NULL, (int*)&jpgctx->nRefCBufferSize,
                        (int*)&jpgctx->nRefYLineStride,
                        (int*)&jpgctx->nRefCLineStride, nAlignValue);
         jpgctx->nDispCBufferSize = jpgctx->nRefCBufferSize;
         jpgctx->nDispYLineStride = jpgctx->nRefYLineStride;
         jpgctx->nDispCLineStride = jpgctx->nRefCLineStride;
         MjpegComputeOffset(pMjpegContext,jpgctx, width, height);
    }
    else
    {
        memset(&mFbmCreateInfo, 0, sizeof(FbmCreateInfo));
        mFbmCreateInfo.nDecoderNeededMiniFrameNum = pFbmInfo->nDecoderNeededMiniFbmNum;
        mFbmCreateInfo.nFrameNum          = pFbmInfo->nDecoderNeededMiniFbmNum;
        mFbmCreateInfo.nWidth             = width;
        mFbmCreateInfo.nHeight            = height;
        mFbmCreateInfo.ePixelFormat       = nRefPixelFormat;
        mFbmCreateInfo.bThumbnailMode     = pMjpegContext->vconfig.bThumbnailMode;
        //mFbmCreateInfo.callback           = pMjpegContext->vconfig.callback;
        //mFbmCreateInfo.pUserData          = pMjpegContext->vconfig.pUserData;
        mFbmCreateInfo.bGpuBufValid       = pMjpegContext->vconfig.bGpuBufValid;
        mFbmCreateInfo.nAlignStride       = pMjpegContext->vconfig.nAlignStride;
        mFbmCreateInfo.nBufferType        = BUF_TYPE_ONLY_REFERENCE;
        mFbmCreateInfo.bProgressiveFlag   = nProgressiveFlag;
        mFbmCreateInfo.bIsSoftDecoderFlag = pMjpegContext->vconfig.bIsSoftDecoderFlag;
        mFbmCreateInfo.memops             = pMjpegContext->vconfig.memops;
        mFbmCreateInfo.veOpsS             = pMjpegContext->vconfig.veOpsS;

        jpgctx->pFbm = FbmCreate(&mFbmCreateInfo, pMjpegContext->pFbmInfo);

        nAlignValue =  FbmGetAlignValue(jpgctx->pFbm);
        GetBufferSize((int)nRefPixelFormat,(int)width,(int)height,
                   (int*)NULL, (int*)&jpgctx->nRefCBufferSize,
                   (int*)&jpgctx->nRefYLineStride,
                   (int*)&jpgctx->nRefCLineStride, nAlignValue);

        if(jpgctx->rotationDegree==1 || jpgctx->rotationDegree==3)
        {
            height = jpgctx->nWidth;
            width = jpgctx->nHeight;
        }
        width >>= pMjpegContext->vconfig.nHorizonScaleDownRatio;
        height >>= pMjpegContext->vconfig.nVerticalScaleDownRatio;

        //lineStride = (width+63) & ~63;
        memset(&mFbmCreateInfo, 0, sizeof(FbmCreateInfo));
        mFbmCreateInfo.nFrameNum          =
        pFbmInfo->nDecoderNeededMiniFbmNumSD + pFbmInfo->nExtraFbmBufferNum;
        mFbmCreateInfo.nDecoderNeededMiniFrameNum = pFbmInfo->nDecoderNeededMiniFbmNum;
        mFbmCreateInfo.nWidth             = width;
        mFbmCreateInfo.nHeight            = height;
        mFbmCreateInfo.ePixelFormat       = nDispPixelFormat;
        mFbmCreateInfo.bThumbnailMode     = pMjpegContext->vconfig.bThumbnailMode;

        #if 1
        if(pMjpegContext->vconfig.bThumbnailMode == 1)
        {
            if(pMjpegContext->vconfig.bRotationEn==1 && pMjpegContext->vconfig.bScaleDownEn==1)
            {
                mFbmCreateInfo.bThumbnailMode = 0;
                mFbmCreateInfo.nFrameNum = 2;
            }
        }
        #endif
        mFbmCreateInfo.bGpuBufValid       = pMjpegContext->vconfig.bGpuBufValid;
        mFbmCreateInfo.nAlignStride       = pMjpegContext->vconfig.nAlignStride;
        mFbmCreateInfo.nBufferType        = BUF_TYPE_ONLY_DISP;
        mFbmCreateInfo.bProgressiveFlag   = nProgressiveFlag;
        mFbmCreateInfo.bIsSoftDecoderFlag = pMjpegContext->vconfig.bIsSoftDecoderFlag;
        mFbmCreateInfo.memops             = pMjpegContext->vconfig.memops;
        mFbmCreateInfo.veOpsS             = pMjpegContext->vconfig.veOpsS;

        jpgctx->pFbmScaleDown = FbmCreate(&mFbmCreateInfo, pMjpegContext->pFbmInfo);

        nAlignValue =  FbmGetAlignValue(jpgctx->pFbmScaleDown);
        GetBufferSize((int)nDispPixelFormat,(int)width,(int)height,
                (int*)NULL, (int*)&jpgctx->nDispCBufferSize,
                (int*)&jpgctx->nDispYLineStride,
                (int*)&jpgctx->nDispCLineStride, nAlignValue);

         if((jpgctx->pFbm==NULL) || (jpgctx->pFbmScaleDown==NULL))
         {
             return VDECODE_RESULT_UNSUPPORTED;
         }
         MjpegComputeOffset(pMjpegContext,jpgctx, width, height);

    }

    if(jpgctx->nDecStramIndex == 0)
    {
        pMjpegContext->pFbm = jpgctx->pFbm;
        pMjpegContext->pFbmScaleDown = jpgctx->pFbmScaleDown;
    }
    else if(jpgctx->nDecStramIndex == 1)
    {
        pMjpegContext->pMinorFbm = jpgctx->pFbm;
        pMjpegContext->pMinorFbmScaleDown = jpgctx->pFbmScaleDown;
    }
    return VDECODE_RESULT_OK;
}

s32 JpegGetSof(JpegDec* jpgctx)
{
    s32         length = 0;
    u8            data_precision = 0;
    s32         ci = 0;
    s32           c = 0;

    INPUT_2BYTES(length);
    INPUT_BYTE(data_precision);

    INPUT_2BYTES(jpgctx->nHeight);
    INPUT_2BYTES(jpgctx->nWidth);

    INPUT_BYTE(jpgctx->nNumComponents);
    length -= 8;

    if (jpgctx->sawSOF)
        return 0;

    if (data_precision != BITS_IN_JSAMPLE)
        return 0;

    if (jpgctx->nNumComponents > MAX_COMPONENTS)
        return 0;

    if (length != (jpgctx->nNumComponents * 3))
        return 0;

    for (ci = 0; ci < jpgctx->nNumComponents; ci++)
    {
        JpegComponentInfo* compptr = &jpgctx->curCompInfo[ci];
        compptr->componentIndex = ci;
        INPUT_BYTE(compptr->componentId);
        INPUT_BYTE(c);
        compptr->hSampFactor = (c >> 4) & 15;
        compptr->vSampFactor = (c     ) & 15;
        INPUT_BYTE(compptr->quantTblNo);
    }

    //jpgctx->bitstream += length;
    jpgctx->sawSOF = 1;
    return 1;
}

s32 JpegGetDqt(JpegDec* jpgctx)
{
    s32 length;
    s32 n, i,prec;
    u32 tmp;

    INPUT_2BYTES(length);
    length -= 2;

    while (length > 0)
    {
        INPUT_BYTE(n);
        prec = n >> 4;
        n &= 0x0F;

        if (n >= NUM_QUANT_TBLS)
            return 0;

        u16 *QtabPtr = jpgctx->QTab[n];

        for (i = 0; i < DCTSIZE2; i++)
        {
            if (prec)
            {
                INPUT_2BYTES(tmp);
            }
            else
            {
                INPUT_BYTE(tmp);
            }
            QtabPtr[i] = (u16) tmp;
        }

        length -= DCTSIZE2+1;
        if (prec)
            length -= DCTSIZE2;
    }

    if (length != 0)
        return 0;

    return 1;
}

s32 JpegGetDht(JpegDec* jpgctx)
{
    s32         length;
    u8             bits[17];
    s32         i;
    s32         index;
    s32         count;
    s32            v;
    s32            code_max;
    JHuffTable*    htblptr;
    s32         nb;
    s32            code;

    INPUT_2BYTES(length);
    length -= 2;
    jpgctx->hasDht = 1;

    while (length > 16)
    {
        INPUT_BYTE(index);
        if (index & 0x10)
        {    /* AC table definition */
            index -= 0x10;
            htblptr = &jpgctx->acHuffTbl[index];
        }
        else
        {    /* DC table definition */
            htblptr = &jpgctx->dcHuffTbl[index];
        }

        if (index < 0 || index >= NUM_HUFF_TBLS)
            return 0;

        bits[0] = 0;
        count = 0;
        for (i = 1; i <= 16; i++)
        {
            htblptr->offset[i-1] = (u8)count;
            INPUT_BYTE(bits[i]);
            count += bits[i];
        }

        length -= 1 + 16;

        if (count > 256 || count > length)
            return 0;

        code_max=0;
        for(i=0;i<count;i++)
        {
            INPUT_BYTE(v);
            if (v > code_max)
                code_max = v;
            htblptr->symbol[i] = (u8)v;
        }

        length -= count;
        {
            code = 0;
            for(i=1;i<=16;i++)
            {
                nb = bits[i];
                htblptr->startCode[i-1] = (u16)code;
                code+=nb;
                code <<= 1;
            }

            for(i=16;i>=1;i--)
            {
                if(bits[i]==0)
                    htblptr->startCode[i-1] = 0xffff;
                else
                    break;
            }
        }
    }

    if (length != 0)
        return 0;

    return 1;
}

s32 JpegGetSos(JpegDec* jpgctx)
{
    s32 length;
    s32 i, ci, n, c, cc;
    JpegComponentInfo* compptr;

    if (!jpgctx->sawSOF)
        return 0;

    INPUT_2BYTES(length);
    INPUT_BYTE(n); /* Number of components */

    if (length != (n * 2 + 6) || n < 1 || n > MAX_COMPS_IN_SCAN)
        return 0;

    jpgctx->nCompsInScan = n;

    for (i = 0; i < n; i++)
    {
        INPUT_BYTE(cc);
        INPUT_BYTE(c);
        for (ci = 0; ci < jpgctx->nNumComponents; ci++)
        {
            compptr = &jpgctx->curCompInfo[ci];
            if (cc == compptr->componentId)
                goto id_found;
        }
        return 0;
id_found:
        jpgctx->sosNbBlocks[i] = compptr->hSampFactor * compptr->vSampFactor;
        jpgctx->seqCompId[i] = compptr->componentId;
        jpgctx->sosHCount[i] = (u8)compptr->hSampFactor;
        jpgctx->sosVCount[i] = (u8)compptr->vSampFactor;

        compptr->dcTblNo = (c >> 4) & 15;
        compptr->acTblNo = (c     ) & 15;
    }

    /* Collect the additional scan parameters Ss, Se, Ah/Al. */
    INPUT_BYTE(c);
    INPUT_BYTE(cc);
    INPUT_BYTE(ci);
    if(c!=0 || cc!=63 || ci!=0)
    {
        return 0;
    }
    jpgctx->nNextRestartNum = 0;

    return 1;
}

s32 JpegGetDri(JpegDec* jpgctx)
{
    s32 length;
    u32 tmp;

    INPUT_2BYTES(length);
    if (length != 4)
        return 0;
    INPUT_2BYTES(tmp);
    jpgctx->nRestartInterval = tmp;

    return 1;
}

s32 JpegReadMarkers(JpegDec* jpgctx)
{
    jpgctx->hasDht = 0;

    for (;;)
    {
        if (jpgctx->unreadMarker == 0)
        {
            if (!jpgctx->sawSOI)
            {
                if (!JpegFirstMarker(jpgctx))
                    return JPEG_MARKER_ERROR;
            }else
            {
                if(!JpegNextMarker(jpgctx))//not found next mark
                    return JPEG_MARKER_ERROR;
            }
        }

        switch(jpgctx->unreadMarker)
        {
        case M_SOI:
            if (!JpegGetSoi(jpgctx))
                return JPEG_MARKER_ERROR;
            break;

        case M_SOF0:                /* Baseline */
        case M_SOF1:                /* Extended sequential, Huffman */
            if (!JpegGetSof(jpgctx))
                return JPEG_MARKER_ERROR;
            break;
            /* Currently unsupported SOFn types */
        case M_SOF2:                /* Progressive, Huffman */
        case M_SOF3:                /* Lossless, Huffman */
        case M_SOF5:                /* Differential sequential, Huffman */
        case M_SOF6:                /* Differential progressive, Huffman */
        case M_SOF7:                /* Differential lossless, Huffman */
        case M_JPG:                 /* Reserved for JPEG extensions */
        case M_SOF9:                /* Extended sequential, arithmetic */
        case M_SOF10:               /* Progressive, arithmetic */
        case M_SOF11:               /* Lossless, arithmetic */
        case M_SOF13:               /* Differential sequential, arithmetic */
        case M_SOF14:               /* Differential progressive, arithmetic */
        case M_SOF15:               /* Differential lossless, arithmetic */
            return JPEG_FORMAT_UNSUPPORT;//?

        case M_SOS:
            if (!JpegGetSos(jpgctx))
                return JPEG_MARKER_ERROR;
            //no break here
        case M_EOI:
            jpgctx->unreadMarker = 0; /* processed the marker */
            return JPEG_PARSER_OK;
        case M_DHT:
            if (!JpegGetDht(jpgctx))
                return JPEG_MARKER_ERROR;
            break;

        case M_DQT:
            if (!JpegGetDqt(jpgctx))
                return JPEG_MARKER_ERROR;
            break;

        case M_DRI:
            if (! JpegGetDri(jpgctx))
                return JPEG_MARKER_ERROR;
            break;

        case M_COM : //skip com
        case M_DAC : //D_ARITH_CODING_NOT_SUPPORTED ?
        case M_APP0:
        case M_APP1: //no need to process EXIF for jpeg decoder
        case M_APP2:
        case M_APP3:
        case M_APP4:
        case M_APP5:
        case M_APP6:
        case M_APP7:
        case M_APP8:
        case M_APP9:
        case M_APP10:
        case M_APP11:
        case M_APP12:
        case M_APP13:
        case M_APP14:
        case M_APP15:
        case M_DNL:
            //! may be some marker need process
            JpegSkipVariable (jpgctx);
            break;

        case M_RST0: /* these are all parameterless */
        case M_RST1:
        case M_RST2:
        case M_RST3:
        case M_RST4:
        case M_RST5:
        case M_RST6:
        case M_RST7:
        case M_TEM:
            break;

        default:  /* must be DHP, EXP, JPGn, or RESn */
            return JPEG_MARKER_ERROR;//?
            //break;
        }
        jpgctx->unreadMarker = 0;
    }
}

s32 JpegDecoderMain(MjpegDecodeContext* pMjpegContext,  JpegDec* jpgctx)
{
    s32 ret;

    jpgctx->sawSOI = 0;
    jpgctx->sawSOF = 0;
    jpgctx->unreadMarker = 0;

    ret = JpegReadMarkers(jpgctx);
    if(JPEG_PARSER_OK != ret)
    {
        return -1;//DEC_ERROR;
    }

    if(!SetJpegFormat(jpgctx))
    {
        return -1;//DEC_ERROR;
    }
    if(jpgctx->pFbm == NULL)
    {
        ret = JpegMallocFrmBuffer(pMjpegContext,jpgctx);
        if(ret !=  VDECODE_RESULT_OK)
        {
            return ret;
        }
    }

    jpgctx->pRefPicture =  FbmRequestBuffer(jpgctx->pFbm);
    if(jpgctx->pRefPicture == NULL)
    {
        return VDECODE_RESULT_NO_FRAME_BUFFER;
    }

    // step: decode the stream data
    jpgctx->nRefPicLumaAddr = jpgctx->pRefPicture->phyYBufAddr+
        FbmGetBufferOffset(jpgctx->pFbm, 1);
    jpgctx->nRefPicChromaAddr = jpgctx->pRefPicture->phyCBufAddr+
        FbmGetBufferOffset(jpgctx->pFbm, 0);

    if((jpgctx->enScaleDownOut==1) ||
        (jpgctx->bYv12OutFlag==1) ||(jpgctx->bNv21OutFlag==1))//1633
    {
        jpgctx->pReconPicture =  FbmRequestBuffer(jpgctx->pFbmScaleDown);
        if(jpgctx->pReconPicture == NULL)
        {
            FbmReturnBuffer(jpgctx->pFbm, jpgctx->pRefPicture, 0);
            jpgctx->pRefPicture = NULL;
            return VDECODE_RESULT_NO_FRAME_BUFFER;
        }
        jpgctx->nReconPicLumaAddr = jpgctx->pReconPicture->phyYBufAddr+
            FbmGetBufferOffset(jpgctx->pFbmScaleDown, 1);
        jpgctx->nReconPicChromaAddr = jpgctx->pReconPicture->phyCBufAddr+
            FbmGetBufferOffset(jpgctx->pFbmScaleDown, 0);
    }
    if(!InitJpegHw(jpgctx, pMjpegContext))
    {
        if(jpgctx->pRefPicture != NULL)
        {
            FbmReturnBuffer(jpgctx->pFbm, jpgctx->pRefPicture, 0);
            jpgctx->pRefPicture = NULL;
        }
        if(jpgctx->pReconPicture != NULL)
        {
            FbmReturnBuffer(jpgctx->pFbmScaleDown, jpgctx->pReconPicture, 0);
            jpgctx->pReconPicture = NULL;
        }
        return -1;//DEC_ERROR;
    }
    if(!JpegHwDec(jpgctx, pMjpegContext))
    {
        if(jpgctx->pRefPicture != NULL)
        {
            FbmReturnBuffer(jpgctx->pFbm, jpgctx->pRefPicture, 0);
            jpgctx->pRefPicture = NULL;
        }
        if(jpgctx->pReconPicture != NULL)
        {
            FbmReturnBuffer(jpgctx->pFbmScaleDown, jpgctx->pReconPicture, 0);
            jpgctx->pReconPicture = NULL;
        }
        return -1;//DEC_ERROR;
    }
    return ret;
}

void MjpegSetDispParams(MjpegDecodeContext* pMjpegContext,
    JpegDec* mJpegDec, VideoPicture* pCurPicture)
{
    pCurPicture->nFrameRate  = pMjpegContext->videoStreamInfo.nFrameRate;
    pCurPicture->nAspectRatio  = 1000;
    pCurPicture->nPts = mJpegDec->nPts;
    pCurPicture->bIsProgressive = 1;

    pCurPicture->nTopOffset    = mJpegDec->nTopOffset;
    pCurPicture->nLeftOffset   = mJpegDec->nLeftOffset;
    pCurPicture->nRightOffset  = mJpegDec->nRightOffset;
    pCurPicture->nBottomOffset = mJpegDec->nBottomOffset;

     if(pMjpegContext->vconfig.bScaleDownEn == 1)
     {
         pCurPicture->nAspectRatio <<= pMjpegContext->vconfig.nHorizonScaleDownRatio;
         pCurPicture->nAspectRatio >>= pMjpegContext->vconfig.nVerticalScaleDownRatio;
     }
}

void MjpegTransformPlanner422To420(MjpegDecodeContext* pMjpegContext,JpegDec*pMjpegDec, u8*Y)
{
    u8* src;
    u8* dst;
    s32 csize;
    s32 widthAlign;
    s32 heightAlign;
    s32 i;
    s32 width;
    s32 height;
    u8* C;
    struct mem_interface *_memops;

    _memops = pMjpegContext->vconfig.memops;

    width = pMjpegDec->nWidth >> pMjpegDec->horizontalScaleRatio;
    height = pMjpegDec->nHeight>> pMjpegDec->verticalScaleRatio;

    widthAlign  = (width  + 15) & ~15;
    heightAlign = (height + 15) & ~15;

    csize = widthAlign * heightAlign;
    src = C = Y + widthAlign * heightAlign;
    dst = src;
    widthAlign = (widthAlign/2 + 7) & ~7;

    for(i=0; i<heightAlign*2; i+=2)
    {
        memcpy(dst, src, widthAlign);
        dst += widthAlign;
        src += (2*widthAlign);
    }

    cdc_mem_flush_cache(_memops, (u8*)C, csize*2);
    return;
}


s32 RotateMjpegBySoft(u8* JpegLumaSrcStartAddr,
    u8* JpegLumaOutStartAddr,s32 rotateAngle,
    s32 width, s32 height, MjpegDecodeContext* pMjpegContext)
{

    s32 x;
    s32 y;
    u32 ysize;
    u32 csize;
    u8* dst;
    u8* src;
    u32 nAlignStride = pMjpegContext->vconfig.nAlignStride;
    JpegDec*        pMjpegDec = NULL;
    struct mem_interface *_memops;

    _memops = pMjpegContext->vconfig.memops;
    pMjpegDec = (JpegDec*)pMjpegContext->pMjpegDec;

    width  = (width+nAlignStride-1) &~ (nAlignStride-1);
    height = (height+nAlignStride-1) &~ (nAlignStride-1);

    ysize  = width*height;
    csize  = ysize/4;

    src = JpegLumaSrcStartAddr;
    dst = JpegLumaOutStartAddr;

    if(rotateAngle == 1)
    {
        for(y=0; y<height; y++)
        {
            for(x=0; x<width; x++)
            {
                dst[x*height+(height-y)-1] = src[y*width+x];
            }
        }

        src = JpegLumaSrcStartAddr + ysize;
        dst = JpegLumaOutStartAddr + ysize;
        width >>= 1;
        height >>= 1;

        if((pMjpegDec->bYv12OutFlag==1) ||
            ((pMjpegDec->bRefDispSameFormat==1) &&
            ((pMjpegDec->eDispFormat==DISP_FORMAT_YU12)||
             (pMjpegDec->eDispFormat==DISP_FORMAT_YV12))))
        {

            for(y=0; y<height; y++)
            {
                for(x=0; x<width; x++)
                {
                    dst[x*height+(height-y)-1] = src[y*width+x];
                }
            }

            src += csize;
            dst += csize;
            for(y=0; y<height; y++)
            {
                for(x=0; x<width; x++)
                {
                    dst[x*height+(height-y)-1] = src[y*width+x];
                }
            }
        }
        else
        {
            for(y=0; y<height; y++)
            {
                for(x=0; x<width; x++)
                {
                    dst[2*x*height+2*height-(2*y+1)-1] = src[2*y*width+2*x];
                    dst[2*x*height+2*height-(2*y)-1] = src[2*y*width+2*x+1];
                }
            }
        }
    }
    else if(rotateAngle == 2)
    {
        for(y=0; y<height; y++)
        {
            for(x=0; x<width; x++)
            {
                dst[(height-y-1)*width+(width-x-1)] = src[y*width+x];
            }
        }

        src = JpegLumaSrcStartAddr + ysize;
        dst = JpegLumaOutStartAddr + ysize;
        width >>= 1;
        height >>= 1;
       if((pMjpegDec->bYv12OutFlag==1) ||
            ((pMjpegDec->bRefDispSameFormat==1) &&
            ((pMjpegDec->eDispFormat==DISP_FORMAT_YU12)||
             (pMjpegDec->eDispFormat==DISP_FORMAT_YV12))))
        {
            for(y=0; y<height; y++)
            {
                for(x=0; x<width; x++)
                {
                    dst[(height-y-1)*width+(width-x-1)] = src[y*width+x];
                }
            }

            src += csize;
            dst += csize;
            for(y=0; y<height; y++)
            {
                for(x=0; x<width; x++)
                {
                    dst[(height-y-1)*width+(width-x-1)] = src[y*width+x];
                }
            }
        }
        else
        {
             for(y=0; y<height; y++)
             {
                 for(x=0; x<width; x++)
                 {
                     dst[2*(height-1-y)*width+2*(width-1-x)] = src[2*y*width+2*x];
                     dst[2*(height-1-y)*width+2*(width-1-x)+1] = src[2*y*width+2*x+1];
                 }
             }
        }
    }
    else if(rotateAngle == 3)
    {
        for(y=0; y<height; y++)
        {
            for(x=0; x<width; x++)
            {
                dst[(width-x-1)*height+y] = src[y*width+x];
            }
        }

        src = JpegLumaSrcStartAddr + ysize;
        dst = JpegLumaOutStartAddr + ysize;
        width >>= 1;
        height >>= 1;
        if((pMjpegDec->bYv12OutFlag==1) ||
            ((pMjpegDec->bRefDispSameFormat==1) &&
            ((pMjpegDec->eDispFormat==DISP_FORMAT_YU12)||
             (pMjpegDec->eDispFormat==DISP_FORMAT_YV12))))
        {
            for(y=0; y<height; y++)
            {
                for(x=0; x<width; x++)
                {
                    dst[(width-x-1)*height+y] = src[y*width+x];
                }
            }

            src += csize;
            dst += csize;
            for(y=0; y<height; y++)
            {
                for(x=0; x<width; x++)
                {
                    dst[(width-x-1)*height+y] = src[y*width+x];
                }
            }
        }
        else
        {
            for(y=0; y<height; y++)
            {
                for(x=0; x<width; x++)
                {
                    dst[2*(width-1-x)*height+2*y] = src[2*y*width+2*x];
                    dst[2*(width-1-x)*height+2*y+1] = src[2*y*width+2*x+1];
                }
            }
        }
    }
    else
    {
        memcpy(dst, src, ysize + csize*2);
    }

    cdc_mem_flush_cache(_memops,(u8*)JpegLumaOutStartAddr, ysize + csize*2);
    return 0;
}

