
/*
 * Copyright (c) 2008-2016 Allwinner Technology Co. Ltd.
 * All rights reserved.
 *
 * File : videoengine.c
 * Description :
 * History :
 *   Author  : jilinglin <jilinglin@allwinnertech.com>
 *   Date    : 2016/04/13
 *   Comment :
 */

#include "videoengine.h"

#include <linux/slab.h>

#include "log.h"
#include "mjpeg.h"
#include "ve_interface.h"
// #include "mjpeg_plus.h"

#define DEBUG_SCLALE_DOWN 0

#define VIDEO_ENGINE_WIDTH_4K 3840
#define VIDEO_ENGINE_HEIGHT_4K 2160

#define TRANSFORM_DEV_TIMEOUT 50  //* ms , tr timeout value

const char* strCodecFormat[] = {
	"MJPEG", "MPEG1", "MPEG2", "MPEG4", "MSMPEG4V1", "MSMPEG4V2",
	"DIVX3", "DIVX4", "DIVX5", "XVID",	"H263",		 "H263SORRENSSON",
	"RXG2",	 "WMV1",  "WMV2",  "WMV3",	"VP6",		 "VP8",
	"VP9",	 "RX",	  "H264",  "H265",	"AVS"};

static void enableVe(VideoEngine* pVideoEngine)
{
	if (pVideoEngine->bIsSoftwareDecoder == 0) {
		ve_enable_ve(pVideoEngine->veOpsS);
	}
}

static void disableVe(VideoEngine* pVideoEngine)
{
	if (pVideoEngine->bIsSoftwareDecoder == 0) {
		ve_disable_ve(pVideoEngine->veOpsS);
	}
}

static DecoderInterface* CreateSpecificDecoder(VideoEngine* p)
{
	DecoderInterface* pInterface = NULL;
	switch (p->videoStreamInfo.eCodecFormat) {
	case VIDEO_CODEC_FORMAT_MJPEG:
		// pInterface = CreateMjpegPlusDecoder();
		pInterface = CreateMjpegDecoder();
		break;
	default:
		loge("format '%x' not support!", p->videoStreamInfo.eCodecFormat);
	}
	return pInterface;
}

static void DestroySpecificDecoder(VideoEngine* p)
{
	switch (p->videoStreamInfo.eCodecFormat) {
	case VIDEO_CODEC_FORMAT_MJPEG:
		// DestroyMjpegPlusDecoder(p->pDecoderInterface);
		DestroyMjpegDecoder(p->pDecoderInterface);
		break;
	default:
		loge("format '%x' not support!", p->videoStreamInfo.eCodecFormat);
	}
}

VideoEngine* VideoEngineCreate(VConfig* pVConfig, VideoStreamInfo* pVideoInfo)
{
	int ret;
	VideoEngine* pEngine;
	pEngine =
		(VideoEngine*)kmalloc(sizeof(VideoEngine), GFP_KERNEL | __GFP_ZERO);
	if (pEngine == NULL) {
		loge("memory alloc fail, VideoEngineCreate() return fail.");
		return NULL;
	}
	memset(pEngine, 0, sizeof(VideoEngine));
	memcpy(&pEngine->vconfig, pVConfig, sizeof(VConfig));
	memcpy(&pEngine->videoStreamInfo, pVideoInfo, sizeof(VideoStreamInfo));

	pEngine->nTrhandle = -1;
	pEngine->nTrChannel = 0;
	pEngine->bInitTrFlag = 0;

	// added by xyliu for dtmb, we should reset ve totallly
	if (pVConfig->bIsTvStream == 1) {
		pEngine->nResetVeMode = RESET_VE_SPECIAL;
	} else {
		pEngine->nResetVeMode = RESET_VE_NORMAL;
	}
	pEngine->veOpsS = pVConfig->veOpsS;
	pEngine->nIcVeVersion = ve_get_ic_version(pEngine->veOpsS);
	pEngine->ndecIpVersion = (uint32_t)(pEngine->nIcVeVersion >> 32);

	//    loge("*** pEngine->nIcVeVersion = %llx, decIpVersion = %x",
	//         (long long unsigned
	//         int)pEngine->nIcVeVersion,pEngine->ndecIpVersion);

	pEngine->pDecoderInterface = CreateSpecificDecoder(pEngine);
	if (pEngine->pDecoderInterface == NULL) {
		loge("unsupported format %s",
			 strCodecFormat[pVideoInfo->eCodecFormat - VIDEO_CODEC_FORMAT_MIN]);
		goto error_exit;
	}

	if (pEngine->bIsSoftwareDecoder == 0) {
		//* reset ve hardware and set up top level registers.
		ve_reset(pEngine->veOpsS);

		if (pEngine->vconfig.eOutputPixelFormat == PIXEL_FORMAT_DEFAULT) {
			pEngine->vconfig.eOutputPixelFormat = PIXEL_FORMAT_YUV_MB32_420;
			if (pEngine->nIcVeVersion >= 0x1667) {
				pEngine->vconfig.eOutputPixelFormat = PIXEL_FORMAT_YV12;
			}
		}
	}

	logd("**************eCtlAfcbMode = %d", pEngine->vconfig.eCtlAfbcMode);

	pEngine->vconfig.bIsSoftDecoderFlag = pEngine->bIsSoftwareDecoder;
	//* call specific function to open decoder.

	enableVe(pEngine);
	ve_set_ddr_mode(pEngine->veOpsS, DDRTYPE_DDR3_32BITS);
	pEngine->fbmInfo.bIs3DStream =
		pEngine->videoStreamInfo.bIs3DStream;  // added by xyliu
	pEngine->fbmInfo.bIsFrameCtsTestFlag =
		pEngine->videoStreamInfo.bIsFrameCtsTestFlag;  // cts
	pEngine->fbmInfo.nExtraFbmBufferNum =
		pEngine->vconfig.nDeInterlaceHoldingFrameBufferNum +
		pEngine->vconfig.nDisplayHoldingFrameBufferNum +
		pEngine->vconfig.nRotateHoldingFrameBufferNum +
		pEngine->vconfig.nDecodeSmoothFrameBufferNum;
	if (pEngine->fbmInfo.nExtraFbmBufferNum > 16)
		pEngine->fbmInfo.nExtraFbmBufferNum = 16;
	if (pEngine->fbmInfo.nExtraFbmBufferNum < 0) {
		pEngine->fbmInfo.nExtraFbmBufferNum = 0;
		logw(" extra fbm buffer == 0 ");
	}
#if DEBUG_SCLALE_DOWN
	{
		pEngine->vconfig.bScaleDownEn = 1;
		pEngine->vconfig.nHorizonScaleDownRatio = 1;
		pEngine->vconfig.nVerticalScaleDownRatio = 1;
	}
#endif
	ret = pEngine->pDecoderInterface->Init(
		pEngine->pDecoderInterface, &pEngine->vconfig,
		&pEngine->videoStreamInfo, &pEngine->fbmInfo);

	if (ret != VDECODE_RESULT_OK) {
		loge("initial specific decoder fail.");
		goto error_exit;
	}
	return pEngine;

error_exit:

	if (pEngine) {
		kfree(pEngine);
	}
	return NULL;
}

void VideoEngineDestroy(VideoEngine* pVideoEngine)
{
	//*warning: must destory by the order: pDecoderInterface, sbm, ve

	//* close specific decoder.

	logd("ji*************VideoEngineDestroy");
	DestroySpecificDecoder(pVideoEngine);

	disableVe(pVideoEngine);

	kfree(pVideoEngine);

	return;
}

int VideoEngineReopen(VideoEngine* pVideoEngine, VConfig* pVConfig,
					  VideoStreamInfo* pVideoInfo)
{
	int ret = 0;
	VConfig* ppVConfig;

	ppVConfig = pVConfig;

	//* close specific decoder.

	loge("ji*************VideoEngineReopen");
	DestroySpecificDecoder(pVideoEngine);
	pVideoEngine->fbmInfo.nValidBufNum = 0;
	pVideoEngine->fbmInfo.pMajorDispFrame = NULL;
	pVideoEngine->fbmInfo.pMajorDecoderFrame = NULL;
	pVideoEngine->fbmInfo.pFbmFirst = NULL;
	pVideoEngine->fbmInfo.pFbmSecond = NULL;
	memset(&pVideoEngine->fbmInfo.pFbmBufInfo, 0, sizeof(FbmBufInfo));

	pVideoEngine->pDecoderInterface = CreateSpecificDecoder(pVideoEngine);
	if (pVideoEngine->pDecoderInterface == NULL) {
		loge("unsupported format %s",
			 strCodecFormat[pVideoInfo->eCodecFormat - VIDEO_CODEC_FORMAT_MIN]);
		goto error_exit;
	}

	if (pVideoEngine->bIsSoftwareDecoder == 0) {
		//* reset ve hardware and set up top level registers.
		ve_reset(pVideoEngine->veOpsS);

		if (pVideoEngine->vconfig.eOutputPixelFormat == PIXEL_FORMAT_DEFAULT) {
			pVideoEngine->vconfig.eOutputPixelFormat =
				PIXEL_FORMAT_YUV_MB32_420;
			if (pVideoEngine->nIcVeVersion >= 0x1667) {
				pVideoEngine->vconfig.eOutputPixelFormat = PIXEL_FORMAT_YV12;
			}
		}
	}

	ret = pVideoEngine->pDecoderInterface->Init(
		pVideoEngine->pDecoderInterface, &pVideoEngine->vconfig,
		&pVideoEngine->videoStreamInfo, &pVideoEngine->fbmInfo);

	if (ret != VDECODE_RESULT_OK) {
		loge("initial specific decoder fail.");
		return -1;
	}
	return 0;
error_exit:
	return -1;
}

void VideoEngineReset(VideoEngine* pVideoEngine)
{
	//	loge("ji*************VideoEngineReset");
	pVideoEngine->pDecoderInterface->Reset(pVideoEngine->pDecoderInterface);

	return;
}

int VideoEngineSetSbm(VideoEngine* pVideoEngine, SbmInterface* pSbm, int nIndex)
{
	int ret;

	//	loge("ji*************VideoEngineSetSbm");

	if (nIndex < 2)
		pVideoEngine->pSbm[nIndex] = pSbm;

	ret = pVideoEngine->pDecoderInterface->SetSbm(
		pVideoEngine->pDecoderInterface, pSbm, nIndex);
	return ret;
}

int VideoEngineGetFbmNum(VideoEngine* pVideoEngine)
{
	int ret;
	loge("ji*************VideoEngineGetFbmNum");
	ret = pVideoEngine->pDecoderInterface->GetFbmNum(
		pVideoEngine->pDecoderInterface);
	return ret;
}

Fbm* VideoEngineGetFbm(VideoEngine* pVideoEngine, int nIndex)
{
	Fbm* pFbm;
	//	loge("ji***********VideoEngineGetFbm");
	pFbm = pVideoEngine->pDecoderInterface->GetFbm(
		pVideoEngine->pDecoderInterface, nIndex);
	return pFbm;
}

int VideoEngineDecode(VideoEngine* pVideoEngine, int bEndOfStream,
					  int bDecodeKeyFrameOnly, int bDropBFrameIfDelay,
					  int64_t nCurrentTimeUs)
{
	int ret;
	//	loge("ji*************VideoEngineDecode");
	ret = pVideoEngine->pDecoderInterface->Decode(
		pVideoEngine->pDecoderInterface, bEndOfStream, bDecodeKeyFrameOnly,
		bDropBFrameIfDelay, nCurrentTimeUs);

	return ret;
}
